"""
Functions that wrap other functions.

.. versionadded:: 1.0.0
"""

from inspect import getfullargspec
import itertools
import time
import typing as t

from typing_extensions import Concatenate, Literal, ParamSpec

import pydash as pyd


__all__ = (
    "after",
    "ary",
    "before",
    "conjoin",
    "curry",
    "curry_right",
    "debounce",
    "delay",
    "disjoin",
    "flip",
    "flow",
    "flow_right",
    "iterated",
    "juxtapose",
    "negate",
    "once",
    "over_args",
    "partial",
    "partial_right",
    "rearg",
    "spread",
    "throttle",
    "unary",
    "wrap",
)

T = t.TypeVar("T")
T1 = t.TypeVar("T1")
T2 = t.TypeVar("T2")
T3 = t.TypeVar("T3")
T4 = t.TypeVar("T4")
T5 = t.TypeVar("T5")
P = ParamSpec("P")


class After(t.Generic[P, T]):
    """Wrap a function in an after context."""

    def __init__(self, func: t.Callable[P, T], n: t.SupportsInt) -> None:
        try:
            n = int(n)
            assert n >= 0
        except (ValueError, TypeError, AssertionError):
            n = 0

        self.n = n
        self.func = func

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> t.Union[T, None]:
        """Return results of :attr:`func` after :attr:`n` calls."""
        self.n -= 1

        if self.n <= 0:
            return self.func(*args, **kwargs)

        return None


class Ary(t.Generic[T]):
    """Wrap a function in an ary context."""

    def __init__(self, func: t.Callable[..., T], n: t.Union[t.SupportsInt, None]) -> None:
        try:
            # Type error would be caught
            n = int(n)  # type: ignore
            assert n >= 0
        except (ValueError, TypeError, AssertionError):
            n = None

        self.n = n
        self.func = func

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> T:
        """
        Return results of :attr:`func` with arguments capped to :attr:`n`.

        Only positional arguments are capped. Any number of keyword arguments are allowed.
        """
        cut_args = args[: self.n] if self.n is not None else args

        return self.func(*cut_args, **kwargs)  # type: ignore


class Before(After, t.Generic[P, T]):
    """Wrap a function in a before context."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> t.Union[T, None]:
        self.n -= 1

        if self.n > 0:
            return self.func(*args, **kwargs)

        return None


class Flow(t.Generic[P, T]):
    """Wrap a function in a flow context."""

    @t.overload
    def __init__(
        self,
        func1: t.Callable[P, T2],
        func2: t.Callable[[T2], T3],
        func3: t.Callable[[T3], T4],
        func4: t.Callable[[T4], T5],
        func5: t.Callable[[T5], T],
        *,
        from_right: bool = True,
    ) -> None:
        ...

    @t.overload
    def __init__(
        self,
        func1: t.Callable[P, T2],
        func2: t.Callable[[T2], T3],
        func3: t.Callable[[T3], T4],
        func4: t.Callable[[T4], T],
        *,
        from_right: bool = True,
    ) -> None:
        ...

    @t.overload
    def __init__(
        self,
        func1: t.Callable[P, T2],
        func2: t.Callable[[T2], T3],
        func3: t.Callable[[T3], T],
        *,
        from_right: bool = True,
    ) -> None:
        ...

    @t.overload
    def __init__(
        self, func1: t.Callable[P, T2], func2: t.Callable[[T2], T], *, from_right: bool = True
    ) -> None:
        ...

    @t.overload
    def __init__(self, func1: t.Callable[P, T], *, from_right: bool = True) -> None:
        ...

    def __init__(self, *funcs, from_right: bool = True) -> None:  # type: ignore
        self.funcs = funcs
        self.from_right = from_right

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """Return results of composing :attr:`funcs`."""
        funcs = list(self.funcs)
        from_index = -1 if self.from_right else 0

        result = None

        while funcs:
            result = funcs.pop(from_index)(*args, **kwargs)
            # Incompatible type in assignements but needed here
            # type safety is ensured from the `__init__` signature
            args = (result,)  # type: ignore
            kwargs = {}  # type: ignore

        # type safety is ensured from the `__init__` signature
        return result  # type: ignore


class Conjoin(t.Generic[T]):
    """Wrap a set of functions in a conjoin context."""

    def __init__(self, *funcs: t.Callable[[T], t.Any]) -> None:
        self.funcs = funcs

    def __call__(self, obj: t.Iterable[T]) -> bool:
        """Return result of conjoin `obj` with :attr:`funcs` predicates."""

        def iteratee(item: T) -> bool:
            return pyd.every(self.funcs, lambda func: func(item))

        return pyd.every(obj, iteratee)


class Curry(t.Generic[T1, T]):
    """Wrap a function in a curry context."""

    def __init__(self, func, arity, args=None, kwargs=None) -> None:
        self.func = func
        self.arity = len(getfullargspec(func).args) if arity is None else arity
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, *args, **kwargs):
        """Store `args` and `kwargs` and call :attr:`func` if we've reached or exceeded the function
        arity."""
        args = self.compose_args(args)
        kwargs.update(self.kwargs)

        if (len(args) + len(kwargs)) >= self.arity:
            args_arity = self.arity - len(kwargs)
            args = args[: (args_arity if args_arity > 0 else 0)]
            curried = self.func(*args, **kwargs)
        else:
            # NOTE: Use self.__class__ so that subclasses will use their own
            # class to generate next iteration of call.
            curried = self.__class__(self.func, self.arity, args, kwargs)

        return curried

    def compose_args(self, new_args):
        """Combine `self.args` with `new_args` and return."""
        return tuple(list(self.args) + list(new_args))


class CurryOne(Curry[T1, T]):
    def __call__(self, arg_one: T1) -> T:
        return super().__call__(arg_one)  # pragma: no cover


class CurryTwo(Curry[T1, CurryOne[T2, T]]):
    @t.overload
    def __call__(self, arg_one: T1) -> CurryOne[T2, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryThree(Curry[T1, CurryTwo[T2, T3, T]]):
    @t.overload
    def __call__(self, arg_one: T1) -> CurryTwo[T2, T3, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2) -> CurryOne[T3, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryFour(Curry[T1, CurryThree[T2, T3, T4, T]]):
    @t.overload
    def __call__(self, arg_one: T1) -> CurryThree[T2, T3, T4, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2) -> CurryTwo[T3, T4, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3) -> CurryOne[T4, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3, arg_four: T4) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryFive(Curry[T1, CurryFour[T2, T3, T4, T5, T]]):
    @t.overload
    def __call__(self, arg_one: T1) -> CurryFour[T2, T3, T4, T5, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2) -> CurryThree[T3, T4, T5, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3) -> CurryTwo[T4, T5, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3, arg_four: T4) -> CurryOne[T5, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T1, arg_two: T2, arg_three: T3, arg_four: T4, arg_five: T5) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryRight(Curry[T5, T]):
    """Wrap a function in a curry-right context."""

    def compose_args(self, new_args):
        return tuple(list(new_args) + list(self.args))


class CurryRightOne(CurryRight[T5, T]):
    def __call__(self, arg_one: T5) -> T:
        return super().__call__(arg_one)  # pragma: no cover


class CurryRightTwo(CurryRight[T5, CurryRightOne[T4, T]]):
    @t.overload
    def __call__(self, arg_one: T5) -> CurryRightOne[T4, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryRightThree(CurryRight[T5, CurryRightTwo[T4, T3, T]]):
    @t.overload
    def __call__(self, arg_one: T5) -> CurryRightTwo[T4, T3, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4) -> CurryRightOne[T3, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4, arg_three: T3) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryRightFour(CurryRight[T5, CurryRightThree[T4, T3, T2, T]]):
    @t.overload
    def __call__(self, arg_one: T5) -> CurryRightThree[T4, T3, T2, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4) -> CurryRightTwo[T3, T2, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4, arg_three: T3) -> CurryRightOne[T2, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4, arg_three: T3, arg_four: T2) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class CurryRightFive(CurryRight[T5, CurryRightFour[T4, T3, T2, T1, T]]):
    @t.overload
    def __call__(self, arg_one: T5) -> CurryRightFour[T4, T3, T2, T1, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4) -> CurryRightThree[T3, T2, T1, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4, arg_three: T3) -> CurryRightTwo[T2, T1, T]:
        ...

    @t.overload
    def __call__(
        self, arg_one: T5, arg_two: T4, arg_three: T3, arg_four: T2
    ) -> CurryRightOne[T1, T]:
        ...

    @t.overload
    def __call__(self, arg_one: T5, arg_two: T4, arg_three: T3, arg_four: T2, arg_five: T1) -> T:
        ...

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)  # pragma: no cover


class Debounce(t.Generic[P, T]):
    """Wrap a function in a debounce context."""

    def __init__(
        self, func: t.Callable[P, T], wait: int, max_wait: t.Union[int, Literal[False]] = False
    ) -> None:
        self.func = func
        self.wait = wait
        self.max_wait = max_wait

        self.last_result: t.Union[T, None] = None

        # Initialize last_* times to be prior to the wait periods so that func
        # is primed to be executed on first call.
        self.last_call = pyd.now() - self.wait
        self.last_execution = pyd.now() - max_wait if pyd.is_number(max_wait) else None

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Execute :attr:`func` if function hasn't been called within last :attr:`wait` milliseconds or
        in last :attr:`max_wait` milliseconds.

        Return results of last successful call.
        """
        present = pyd.now()

        if (present - self.last_call) >= self.wait or (
            self.max_wait and (present - self.last_execution) >= self.max_wait  # type: ignore
        ):
            self.last_result = self.func(*args, **kwargs)
            self.last_execution = present

        self.last_call = present

        # It will be set after first call, cannot be `None` anymore
        return self.last_result  # type: ignore


class Disjoin(t.Generic[T]):
    """Wrap a set of functions in a disjoin context."""

    def __init__(self, *funcs: t.Callable[[T], t.Any]) -> None:
        self.funcs = funcs

    def __call__(self, obj: t.Iterable[T]) -> bool:
        """Return result of disjoin `obj` with :attr:`funcs` predicates."""

        def iteratee(item: T) -> bool:
            return pyd.some(self.funcs, lambda func: func(item))

        return pyd.some(obj, iteratee)


class Flip(object):
    """Wrap a function in a flip context."""

    def __init__(self, func: t.Callable) -> None:
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*reversed(args), **kwargs)


class Iterated(t.Generic[T]):
    """Wrap a function in an iterated context."""

    def __init__(self, func: t.Callable[[T], T]) -> None:
        self.func = func

    def _iteration(self, initial: T) -> t.Iterator[T]:
        """Iterator that composing :attr:`func` with itself."""
        value = initial
        while True:
            value = self.func(value)
            yield value

    def __call__(self, initial: T, n: int) -> T:
        """Return value of calling :attr:`func` `n` times using `initial` as seed value."""
        value = initial
        iteration = self._iteration(value)

        for _ in range(n):
            value = next(iteration)

        return value


class Juxtapose(t.Generic[P, T]):
    """Wrap a function in a juxtapose context."""

    def __init__(self, *funcs: t.Callable[P, T]) -> None:
        self.funcs = funcs

    def __call__(self, *objs: P.args, **kwargs: P.kwargs) -> t.List[T]:
        return [func(*objs, **kwargs) for func in self.funcs]


class OverArgs(object):
    """Wrap a function in an over_args context."""

    def __init__(self, func: t.Callable, *transforms: t.Callable) -> None:
        self.func = func
        self.transforms = pyd.flatten(transforms)

    def __call__(self, *args):
        args = (self.transforms[idx](args) for idx, args in enumerate(args))
        return self.func(*args)


class Negate(t.Generic[P]):
    """Wrap a function in a negate context."""

    def __init__(self, func: t.Callable[P, t.Any]) -> None:
        self.func = func

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        """Return negated results of calling :attr:`func`."""
        return not self.func(*args, **kwargs)


class Once(t.Generic[P, T]):
    """Wrap a function in a once context."""

    def __init__(self, func: t.Callable[P, T]) -> None:
        self.func = func
        self.result: t.Union[T, None] = None
        self.called = False

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """Return results from the first call of :attr:`func`."""
        if not self.called:
            self.result = self.func(*args, **kwargs)
            self.called = True

        # At this point the result will be set, cannot be `None` anymore
        return self.result  # type: ignore


class Partial(t.Generic[T]):
    """Wrap a function in a partial context."""

    def __init__(
        self, func: t.Callable[..., T], args: t.Any, kwargs: t.Any = None, from_right: bool = False
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs or {}
        self.from_right = from_right

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> T:
        """
        Return results from :attr:`func` with :attr:`args` + `args`.

        Apply arguments from left or right depending on :attr:`from_right`.
        """
        if self.from_right:
            args = itertools.chain(args, self.args)  # type: ignore
        else:
            args = itertools.chain(self.args, args)  # type: ignore

        kwargs = {**self.kwargs, **kwargs}

        return self.func(*args, **kwargs)


class Rearg(t.Generic[P, T]):
    """Wrap a function in a rearg context."""

    def __init__(self, func: t.Callable[P, T], *indexes: int) -> None:
        self.func = func

        # Index `indexes` by the index value, so we can do a lookup mapping by walking the function
        # arguments.
        self.indexes = {
            src_index: dest_index for dest_index, src_index in enumerate(pyd.flatten(indexes))
        }

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """Return results from :attr:`func` using rearranged arguments."""
        reargs = {}
        rest = []

        # Walk arguments to ensure each one is added to the final argument list.
        for src_index, arg in enumerate(args):
            # NOTE: dest_index will range from 0 to len(indexes).
            dest_index = self.indexes.get(src_index)

            if dest_index is not None:
                # Remap argument index.
                reargs[dest_index] = arg
            else:
                # Argumnet index is not contained in `indexes` so stick in the back.
                rest.append(arg)

        args = itertools.chain((reargs[key] for key in sorted(reargs)), rest)  # type: ignore

        return self.func(*args, **kwargs)


class Spread(t.Generic[T]):
    """Wrap a function in a spread context."""

    def __init__(self, func: t.Callable[..., T]) -> None:
        self.func = func

    def __call__(self, args: t.Iterable) -> T:
        """Return results from :attr:`func` using array of `args` provided."""
        return self.func(*args)


class Throttle(t.Generic[P, T]):
    """Wrap a function in a throttle context."""

    def __init__(self, func: t.Callable[P, T], wait: int) -> None:
        self.func = func
        self.wait = wait

        self.last_result: t.Union[T, None] = None
        self.last_execution = pyd.now() - self.wait

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Execute :attr:`func` if function hasn't been called within last :attr:`wait` milliseconds.

        Return results of last successful call.
        """
        present = pyd.now()

        if (present - self.last_execution) >= self.wait:
            self.last_result = self.func(*args, **kwargs)
            self.last_execution = present

        # The last result will be filled on first execution, so it is always `T`
        return self.last_result  # type: ignore


def after(func: t.Callable[P, T], n: t.SupportsInt) -> After[P, T]:
    """
    Creates a function that executes `func`, with the arguments of the created function, only after
    being called `n` times.

    Args:
        func: Function to execute.
        n: Number of times `func` must be called before it is executed.

    Returns:
        Function wrapped in an :class:`After` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> after_func = after(func, 3)
        >>> after_func(1, 2, 3)
        >>> after_func(1, 2, 3)
        >>> after_func(1, 2, 3)
        (1, 2, 3)
        >>> after_func(4, 5, 6)
        (4, 5, 6)

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `func` first.
    """
    return After(func, n)


def ary(func: t.Callable[..., T], n: t.Union[t.SupportsInt, None]) -> Ary[T]:
    """
    Creates a function that accepts up to `n` arguments ignoring any additional arguments. Only
    positional arguments are capped. All keyword arguments are allowed through.

    Args:
        func: Function to cap arguments for.
        n: Number of arguments to accept.

    Returns:
        Function wrapped in an :class:`Ary` context.

    Example:

        >>> func = lambda a, b, c=0, d=5: (a, b, c, d)
        >>> ary_func = ary(func, 2)
        >>> ary_func(1, 2, 3, 4, 5, 6)
        (1, 2, 0, 5)
        >>> ary_func(1, 2, 3, 4, 5, 6, c=10, d=20)
        (1, 2, 10, 20)

    .. versionadded:: 3.0.0
    """
    return Ary(func, n)


def before(func: t.Callable[P, T], n: t.SupportsInt) -> Before[P, T]:
    """
    Creates a function that executes `func`, with the arguments of the created function, until it
    has been called `n` times.

    Args:
        func: Function to execute.
        n: Number of times `func` may be executed.

    Returns:
        Function wrapped in an :class:`Before` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> before_func = before(func, 3)
        >>> before_func(1, 2, 3)
        (1, 2, 3)
        >>> before_func(1, 2, 3)
        (1, 2, 3)
        >>> before_func(1, 2, 3)
        >>> before_func(1, 2, 3)

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `func` first.
    """
    return Before(func, n)


def conjoin(*funcs: t.Callable[[T], t.Any]) -> t.Callable[[t.Iterable[T]], bool]:
    """
    Creates a function that composes multiple predicate functions into a single predicate that tests
    whether **all** elements of an object pass each predicate.

    Args:
        *funcs: Function(s) to conjoin.

    Returns:
        Function(s) wrapped in a :class:`Conjoin` context.

    Example:

        >>> conjoiner = conjoin(lambda x: isinstance(x, int), lambda x: x > 3)
        >>> conjoiner([1, 2, 3])
        False
        >>> conjoiner([1.0, 2, 1])
        False
        >>> conjoiner([4.0, 5, 6])
        False
        >>> conjoiner([4, 5, 6])
        True

    .. versionadded:: 2.0.0
    """
    return Conjoin(*funcs)


@t.overload
def curry(func: t.Callable[[T1], T], arity: t.Union[int, None] = None) -> CurryOne[T1, T]:
    ...


@t.overload
def curry(func: t.Callable[[T1, T2], T], arity: t.Union[int, None] = None) -> CurryTwo[T1, T2, T]:
    ...


@t.overload
def curry(
    func: t.Callable[[T1, T2, T3], T], arity: t.Union[int, None] = None
) -> CurryThree[T1, T2, T3, T]:
    ...


@t.overload
def curry(
    func: t.Callable[[T1, T2, T3, T4], T], arity: t.Union[int, None] = None
) -> CurryFour[T1, T2, T3, T4, T]:
    ...


@t.overload
def curry(
    func: t.Callable[[T1, T2, T3, T4, T5], T], arity: t.Union[int, None] = None
) -> CurryFive[T1, T2, T3, T4, T5, T]:
    ...


def curry(func, arity=None):
    """
    Creates a function that accepts one or more arguments of `func` that when invoked either
    executes `func` returning its result (if all `func` arguments have been provided) or returns a
    function that accepts one or more of the remaining `func` arguments, and so on.

    Args:
        func: Function to curry.
        arity: Number of function arguments that can be accepted by curried
            function. Default is to use the number of arguments that are accepted by `func`.

    Returns:
        Function wrapped in a :class:`Curry` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> currier = curry(func)
        >>> currier = currier(1)
        >>> assert isinstance(currier, Curry)
        >>> currier = currier(2)
        >>> assert isinstance(currier, Curry)
        >>> currier = currier(3)
        >>> currier
        (1, 2, 3)

    .. versionadded:: 1.0.0
    """
    return Curry(func, arity)


@t.overload
def curry_right(
    func: t.Callable[[T1], T], arity: t.Union[int, None] = None
) -> CurryRightOne[T1, T]:
    ...


@t.overload
def curry_right(
    func: t.Callable[[T1, T2], T], arity: t.Union[int, None] = None
) -> CurryRightTwo[T2, T1, T]:
    ...


@t.overload
def curry_right(
    func: t.Callable[[T1, T2, T3], T], arity: t.Union[int, None] = None
) -> CurryRightThree[T3, T2, T1, T]:
    ...


@t.overload
def curry_right(
    func: t.Callable[[T1, T2, T3, T4], T], arity: t.Union[int, None] = None
) -> CurryRightFour[T4, T3, T2, T1, T]:
    ...


@t.overload
def curry_right(func: t.Callable[[T1, T2, T3, T4, T5], T]) -> CurryRightFive[T5, T4, T3, T2, T1, T]:
    ...


def curry_right(func, arity=None):
    """
    This method is like :func:`curry` except that arguments are applied to `func` in the manner of
    :func:`partial_right` instead of :func:`partial`.

    Args:
        func: Function to curry.
        arity: Number of function arguments that can be accepted by curried
            function. Default is to use the number of arguments that are accepted by `func`.

    Returns:
        Function wrapped in a :class:`CurryRight` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> currier = curry_right(func)
        >>> currier = currier(1)
        >>> assert isinstance(currier, CurryRight)
        >>> currier = currier(2)
        >>> assert isinstance(currier, CurryRight)
        >>> currier = currier(3)
        >>> currier
        (3, 2, 1)

    .. versionadded:: 1.1.0
    """
    return CurryRight(func, arity)


def debounce(
    func: t.Callable[P, T], wait: int, max_wait: t.Union[int, Literal[False]] = False
) -> Debounce[P, T]:
    """
    Creates a function that will delay the execution of `func` until after `wait` milliseconds have
    elapsed since the last time it was invoked. Subsequent calls to the debounced function will
    return the result of the last `func` call.

    Args:
        func: Function to execute.
        wait: Milliseconds to wait before executing `func`.
        max_wait (optional): Maximum time to wait before executing `func`.

    Returns:
        Function wrapped in a :class:`Debounce` context.

    .. versionadded:: 1.0.0
    """
    return Debounce(func, wait, max_wait=max_wait)


def delay(func: t.Callable[P, T], wait: int, *args: "P.args", **kwargs: "P.kwargs") -> T:
    """
    Executes the `func` function after `wait` milliseconds. Additional arguments will be provided to
    `func` when it is invoked.

    Args:
        func: Function to execute.
        wait: Milliseconds to wait before executing `func`.
        *args: Arguments to pass to `func`.
        **kwargs: Keyword arguments to pass to `func`.

    Returns:
        Return from `func`.

    .. versionadded:: 1.0.0
    """
    time.sleep(wait / 1000.0)
    return func(*args, **kwargs)


def disjoin(*funcs: t.Callable[[T], t.Any]) -> Disjoin[T]:
    """
    Creates a function that composes multiple predicate functions into a single predicate that tests
    whether **any** elements of an object pass each predicate.

    Args:
        *funcs: Function(s) to disjoin.

    Returns:
        Function(s) wrapped in a :class:`Disjoin` context.

    Example:

        >>> disjoiner = disjoin(lambda x: isinstance(x, float),\
                                lambda x: isinstance(x, int))
        >>> disjoiner([1, '2', '3'])
        True
        >>> disjoiner([1.0, '2', '3'])
        True
        >>> disjoiner(['1', '2', '3'])
        False

    .. versionadded:: 2.0.0
    """
    return Disjoin(*funcs)


@t.overload
def flip(func: t.Callable[[T1, T2, T3, T4, T5], T]) -> t.Callable[[T5, T4, T3, T2, T1], T]:
    ...


@t.overload
def flip(func: t.Callable[[T1, T2, T3, T4], T]) -> t.Callable[[T4, T3, T2, T1], T]:
    ...


@t.overload
def flip(func: t.Callable[[T1, T2, T3], T]) -> t.Callable[[T3, T2, T1], T]:
    ...


@t.overload
def flip(func: t.Callable[[T1, T2], T]) -> t.Callable[[T2, T1], T]:
    ...


@t.overload
def flip(func: t.Callable[[T1], T]) -> t.Callable[[T1], T]:
    ...


def flip(func: t.Callable) -> t.Callable:
    """
    Creates a function that invokes the method with arguments reversed.

    Args:
        func: Function to flip arguments for.

    Returns:
        Function wrapped in a :class:`Flip` context.

    Example:

        >>> flipped = flip(lambda *args: args)
        >>> flipped(1, 2, 3, 4)
        (4, 3, 2, 1)
        >>> flipped = flip(lambda *args: [i * 2 for i in args])
        >>> flipped(1, 2, 3, 4)
        [8, 6, 4, 2]

    .. versionadded:: 4.0.0
    """
    return Flip(func)


@t.overload
def flow(
    func1: t.Callable[P, T2],
    func2: t.Callable[[T2], T3],
    func3: t.Callable[[T3], T4],
    func4: t.Callable[[T4], T5],
    func5: t.Callable[[T5], T],
) -> Flow[P, T]:
    ...


@t.overload
def flow(
    func1: t.Callable[P, T2],
    func2: t.Callable[[T2], T3],
    func3: t.Callable[[T3], T4],
    func4: t.Callable[[T4], T],
) -> Flow[P, T]:
    ...


@t.overload
def flow(
    func1: t.Callable[P, T2],
    func2: t.Callable[[T2], T3],
    func3: t.Callable[[T3], T],
) -> Flow[P, T]:
    ...


@t.overload
def flow(func1: t.Callable[P, T2], func2: t.Callable[[T2], T]) -> Flow[P, T]:
    ...


@t.overload
def flow(func1: t.Callable[P, T]) -> Flow[P, T]:
    ...


def flow(*funcs):
    """
    Creates a function that is the composition of the provided functions, where each successive
    invocation is supplied the return value of the previous. For example, composing the functions
    ``f()``, ``g()``, and ``h()`` produces ``h(g(f()))``.

    Args:
        *funcs: Function(s) to compose.

    Returns:
        Function(s) wrapped in a :class:`Flow` context.

    Example:

        >>> mult_5 = lambda x: x * 5
        >>> div_10 = lambda x: x / 10.0
        >>> pow_2 = lambda x: x ** 2
        >>> ops = flow(sum, mult_5, div_10, pow_2)
        >>> ops([1, 2, 3, 4])
        25.0

    .. versionadded:: 2.0.0

    .. versionchanged:: 2.3.1
        Added :func:`pipe` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``pipe``.
    """
    return Flow(*funcs, from_right=False)


@t.overload
def flow_right(
    func5: t.Callable[[T4], T],
    func4: t.Callable[[T3], T4],
    func3: t.Callable[[T2], T3],
    func2: t.Callable[[T1], T2],
    func1: t.Callable[P, T1],
) -> Flow[P, T]:
    ...


@t.overload
def flow_right(
    func4: t.Callable[[T3], T],
    func3: t.Callable[[T2], T3],
    func2: t.Callable[[T1], T2],
    func1: t.Callable[P, T1],
) -> Flow[P, T]:
    ...


@t.overload
def flow_right(
    func3: t.Callable[[T2], T],
    func2: t.Callable[[T1], T2],
    func1: t.Callable[P, T1],
) -> Flow[P, T]:
    ...


@t.overload
def flow_right(func2: t.Callable[[T1], T], func1: t.Callable[P, T1]) -> Flow[P, T]:
    ...


@t.overload
def flow_right(func1: t.Callable[P, T]) -> Flow[P, T]:
    ...


def flow_right(*funcs):
    """
    This function is like :func:`flow` except that it creates a function that invokes the provided
    functions from right to left. For example, composing the functions ``f()``, ``g()``, and ``h()``
    produces ``f(g(h()))``.

    Args:
        *funcs: Function(s) to compose.

    Returns:
        Function(s) wrapped in a :class:`Flow` context.

    Example:

        >>> mult_5 = lambda x: x * 5
        >>> div_10 = lambda x: x / 10.0
        >>> pow_2 = lambda x: x ** 2
        >>> ops = flow_right(mult_5, div_10, pow_2, sum)
        >>> ops([1, 2, 3, 4])
        50.0

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0
        Added :func:`flow_right` and made :func:`compose` an alias.

    .. versionchanged:: 2.3.1
        Added :func:`pipe_right` as alias.

    .. versionchanged:: 4.0.0
        Removed aliases ``pipe_right`` and ``compose``.
    """
    return Flow(*funcs, from_right=True)


def iterated(func: t.Callable[[T], T]) -> Iterated[T]:
    """
    Creates a function that is composed with itself. Each call to the iterated function uses the
    previous function call's result as input. Returned :class:`Iterated` instance can be called with
    ``(initial, n)`` where `initial` is the initial value to seed `func` with and `n` is the number
    of times to call `func`.

    Args:
        func: Function to iterate.

    Returns:
        Function wrapped in a :class:`Iterated` context.

    Example:

        >>> doubler = iterated(lambda x: x * 2)
        >>> doubler(4, 5)
        128
        >>> doubler(3, 9)
        1536

    .. versionadded:: 2.0.0
    """
    return Iterated(func)


def juxtapose(*funcs: t.Callable[P, T]) -> Juxtapose[P, T]:
    """
    Creates a function whose return value is a list of the results of calling each `funcs` with the
    supplied arguments.

    Args:
        *funcs: Function(s) to juxtapose.

    Returns:
        Function wrapped in a :class:`Juxtapose` context.

    Example:

        >>> double = lambda x: x * 2
        >>> triple = lambda x: x * 3
        >>> quadruple = lambda x: x * 4
        >>> juxtapose(double, triple, quadruple)(5)
        [10, 15, 20]


    .. versionadded:: 2.0.0
    """
    return Juxtapose(*funcs)


def negate(func: t.Callable[P, t.Any]) -> Negate[P]:
    """
    Creates a function that negates the result of the predicate `func`. The `func` function is
    executed with the arguments of the created function.

    Args:
        func: Function to negate execute.

    Returns:
        Function wrapped in a :class:`Negate` context.

    Example:

        >>> not_is_number = negate(lambda x: isinstance(x, (int, float)))
        >>> not_is_number(1)
        False
        >>> not_is_number('1')
        True

    .. versionadded:: 1.1.0
    """
    return Negate(func)


def once(func: t.Callable[P, T]) -> Once[P, T]:
    """
    Creates a function that is restricted to execute `func` once. Repeat calls to the function will
    return the value of the first call.

    Args:
        func: Function to execute.

    Returns:
        Function wrapped in a :class:`Once` context.

    Example:

        >>> oncer = once(lambda *args: args[0])
        >>> oncer(5)
        5
        >>> oncer(6)
        5

    .. versionadded:: 1.0.0
    """
    return Once(func)


@t.overload
def over_args(
    func: t.Callable[[T1, T2, T3, T4, T5], T],
    transform_one: t.Callable[[T1], T1],
    transform_two: t.Callable[[T2], T2],
    transform_three: t.Callable[[T3], T3],
    transform_four: t.Callable[[T4], T4],
    transform_five: t.Callable[[T5], T5],
) -> t.Callable[[T1, T2, T3, T4, T5], T]:
    ...


@t.overload
def over_args(
    func: t.Callable[[T1, T2, T3, T4], T],
    transform_one: t.Callable[[T1], T1],
    transform_two: t.Callable[[T2], T2],
    transform_three: t.Callable[[T3], T3],
    transform_four: t.Callable[[T4], T4],
) -> t.Callable[[T1, T2, T3, T4], T]:
    ...


@t.overload
def over_args(
    func: t.Callable[[T1, T2, T3], T],
    transform_one: t.Callable[[T1], T1],
    transform_two: t.Callable[[T2], T2],
    transform_three: t.Callable[[T3], T3],
) -> t.Callable[[T1, T2, T3], T]:
    ...


@t.overload
def over_args(
    func: t.Callable[[T1, T2], T],
    transform_one: t.Callable[[T1], T1],
    transform_two: t.Callable[[T2], T2],
) -> t.Callable[[T1, T2], T]:
    ...


@t.overload
def over_args(
    func: t.Callable[[T1], T],
    transform_one: t.Callable[[T1], T1],
) -> t.Callable[[T1], T]:
    ...


def over_args(func: t.Callable, *transforms: t.Callable) -> t.Callable:  # type: ignore
    """
    Creates a function that runs each argument through a corresponding transform function.

    Args:
        func: Function to wrap.
        *transforms: Functions to transform arguments, specified as individual functions
            or lists of functions.

    Returns:
        Function wrapped in a :class:`OverArgs` context.

    Example:

        >>> squared = lambda x: x ** 2
        >>> double = lambda x: x * 2
        >>> modder = over_args(lambda x, y: [x, y], squared, double)
        >>> modder(5, 10)
        [25, 20]

    .. versionadded:: 3.3.0

    .. versionchanged:: 4.0.0
        Renamed from ``mod_args`` to ``over_args``.
    """
    return OverArgs(func, *transforms)


def partial(func: t.Callable[..., T], *args: t.Any, **kwargs: t.Any) -> Partial[T]:
    """
    Creates a function that, when called, invokes `func` with any additional partial arguments
    prepended to those provided to the new function.

    Args:
        func: Function to execute.
        *args: Partial arguments to prepend to function call.
        **kwargs: Partial keyword arguments to bind to function call.

    Returns:
        Function wrapped in a :class:`Partial` context.

    Example:

        >>> dropper = partial(lambda array, n: array[n:], [1, 2, 3, 4])
        >>> dropper(2)
        [3, 4]
        >>> dropper(1)
        [2, 3, 4]
        >>> myrest = partial(lambda array, n: array[n:], n=1)
        >>> myrest([1, 2, 3, 4])
        [2, 3, 4]

    .. versionadded:: 1.0.0
    """
    return Partial(func, args, kwargs)


def partial_right(func: t.Callable[..., T], *args: t.Any, **kwargs: t.Any) -> Partial[T]:
    """
    This method is like :func:`partial` except that partial arguments are appended to those provided
    to the new function.

    Args:
        func: Function to execute.
        *args: Partial arguments to append to function call.
        **kwargs: Partial keyword arguments to bind to function call.

    Returns:
        Function wrapped in a :class:`Partial` context.

    Example:

        >>> myrest = partial_right(lambda array, n: array[n:], 1)
        >>> myrest([1, 2, 3, 4])
        [2, 3, 4]

    .. versionadded:: 1.0.0
    """
    return Partial(func, args, kwargs, from_right=True)


def rearg(func: t.Callable[P, T], *indexes: int) -> Rearg[P, T]:
    """
    Creates a function that invokes `func` with arguments arranged according to the specified
    indexes where the argument value at the first index is provided as the first argument, the
    argument value at the second index is provided as the second argument, and so on.

    Args:
        func: Function to rearrange arguments for.
        *indexes: The arranged argument indexes.

    Returns:
        Function wrapped in a :class:`Rearg` context.

    Example:

        >>> jumble = rearg(lambda *args: args, 1, 2, 3)
        >>> jumble(1, 2, 3)
        (2, 3, 1)
        >>> jumble('a', 'b', 'c', 'd', 'e')
        ('b', 'c', 'd', 'a', 'e')

    .. versionadded:: 3.0.0
    """
    return Rearg(func, *indexes)


def spread(func: t.Callable[..., T]) -> Spread[T]:
    """
    Creates a function that invokes `func` with the array of arguments provided to the created
    function.

    Args:
        func: Function to spread.

    Returns:
        Function wrapped in a :class:`Spread` context.

    Example:

        >>> greet = spread(lambda *people: 'Hello ' + ', '.join(people) + '!')
        >>> greet(['Mike', 'Don', 'Leo'])
        'Hello Mike, Don, Leo!'

    .. versionadded:: 3.1.0
    """
    return Spread(func)


def throttle(func: t.Callable[P, T], wait: int) -> Throttle[P, T]:
    """
    Creates a function that, when executed, will only call the `func` function at most once per
    every `wait` milliseconds. Subsequent calls to the throttled function will return the result of
    the last `func` call.

    Args:
        func: Function to throttle.
        wait: Milliseconds to wait before calling `func` again.

    Returns:
        Results of last `func` call.

    .. versionadded:: 1.0.0
    """
    return Throttle(func, wait)


def unary(func: t.Callable[..., T]) -> Ary[T]:
    """
    Creates a function that accepts up to one argument, ignoring any additional arguments.

    Args:
        func: Function to cap arguments for.

    Returns:
        Function wrapped in an :class:`Ary` context.

    Example:

        >>> func = lambda a, b=1, c=0, d=5: (a, b, c, d)
        >>> unary_func = unary(func)
        >>> unary_func(1, 2, 3, 4, 5, 6)
        (1, 1, 0, 5)
        >>> unary_func(1, 2, 3, 4, 5, 6, b=0, c=10, d=20)
        (1, 0, 10, 20)

    .. versionadded:: 4.0.0
    """
    return Ary(func, 1)


def wrap(value: T1, func: t.Callable[Concatenate[T1, P], T]) -> Partial[T]:
    """
    Creates a function that provides value to the wrapper function as its first argument. Additional
    arguments provided to the function are appended to those provided to the wrapper function.

    Args:
        value: Value provided as first argument to function call.
        func: Function to execute.

    Returns:
        Function wrapped in a :class:`Partial` context.

    Example:

        >>> wrapper = wrap('hello', lambda *args: args)
        >>> wrapper(1, 2)
        ('hello', 1, 2)

    .. versionadded:: 1.0.0
    """
    return Partial(func, (value,))



"""Python port of Lo-Dash."""

__version__ = "7.0.6"

from .arrays import (
    chunk,
    compact,
    concat,
    difference,
    difference_by,
    difference_with,
    drop,
    drop_right,
    drop_right_while,
    drop_while,
    duplicates,
    fill,
    find_index,
    find_last_index,
    flatten,
    flatten_deep,
    flatten_depth,
    from_pairs,
    head,
    index_of,
    initial,
    intercalate,
    interleave,
    intersection,
    intersection_by,
    intersection_with,
    intersperse,
    last,
    last_index_of,
    mapcat,
    nth,
    pop,
    pull,
    pull_all,
    pull_all_by,
    pull_all_with,
    pull_at,
    push,
    remove,
    reverse,
    shift,
    slice_,
    sort,
    sorted_index,
    sorted_index_by,
    sorted_index_of,
    sorted_last_index,
    sorted_last_index_by,
    sorted_last_index_of,
    sorted_uniq,
    sorted_uniq_by,
    splice,
    split_at,
    tail,
    take,
    take_right,
    take_right_while,
    take_while,
    union,
    union_by,
    union_with,
    uniq,
    uniq_by,
    uniq_with,
    unshift,
    unzip,
    unzip_with,
    without,
    xor,
    xor_by,
    xor_with,
    zip_,
    zip_object,
    zip_object_deep,
    zip_with,
)
from .chaining import _Dash, chain, tap, thru
from .collections import (
    at,
    count_by,
    every,
    filter_,
    find,
    find_last,
    flat_map,
    flat_map_deep,
    flat_map_depth,
    for_each,
    for_each_right,
    group_by,
    includes,
    invoke_map,
    key_by,
    map_,
    nest,
    order_by,
    partition,
    pluck,
    reduce_,
    reduce_right,
    reductions,
    reductions_right,
    reject,
    sample,
    sample_size,
    shuffle,
    size,
    some,
    sort_by,
)
from .exceptions import InvalidMethod
from .functions import (
    after,
    ary,
    before,
    conjoin,
    curry,
    curry_right,
    debounce,
    delay,
    disjoin,
    flip,
    flow,
    flow_right,
    iterated,
    juxtapose,
    negate,
    once,
    over_args,
    partial,
    partial_right,
    rearg,
    spread,
    throttle,
    unary,
    wrap,
)
from .numerical import (
    add,
    ceil,
    clamp,
    divide,
    floor,
    max_,
    max_by,
    mean,
    mean_by,
    median,
    min_,
    min_by,
    moving_mean,
    multiply,
    power,
    round_,
    scale,
    slope,
    std_deviation,
    subtract,
    sum_,
    sum_by,
    transpose,
    variance,
    zscore,
)
from .objects import (
    assign,
    assign_with,
    callables,
    clone,
    clone_deep,
    clone_deep_with,
    clone_with,
    defaults,
    defaults_deep,
    find_key,
    find_last_key,
    for_in,
    for_in_right,
    get,
    has,
    invert,
    invert_by,
    invoke,
    keys,
    map_keys,
    map_values,
    map_values_deep,
    merge,
    merge_with,
    omit,
    omit_by,
    parse_int,
    pick,
    pick_by,
    rename_keys,
    set_,
    set_with,
    to_boolean,
    to_dict,
    to_integer,
    to_list,
    to_number,
    to_pairs,
    to_string,
    transform,
    unset,
    update,
    update_with,
    values,
)
from .predicates import (
    eq,
    gt,
    gte,
    in_range,
    is_associative,
    is_blank,
    is_boolean,
    is_builtin,
    is_date,
    is_decreasing,
    is_dict,
    is_empty,
    is_equal,
    is_equal_with,
    is_error,
    is_even,
    is_float,
    is_function,
    is_increasing,
    is_indexed,
    is_instance_of,
    is_integer,
    is_iterable,
    is_json,
    is_list,
    is_match,
    is_match_with,
    is_monotone,
    is_nan,
    is_negative,
    is_none,
    is_number,
    is_object,
    is_odd,
    is_positive,
    is_reg_exp,
    is_set,
    is_strictly_decreasing,
    is_strictly_increasing,
    is_string,
    is_tuple,
    is_zero,
    lt,
    lte,
)
from .strings import (
    camel_case,
    capitalize,
    chars,
    chop,
    chop_right,
    clean,
    count_substr,
    deburr,
    decapitalize,
    ends_with,
    ensure_ends_with,
    ensure_starts_with,
    escape,
    escape_reg_exp,
    has_substr,
    human_case,
    insert_substr,
    join,
    kebab_case,
    lines,
    lower_case,
    lower_first,
    number_format,
    pad,
    pad_end,
    pad_start,
    pascal_case,
    predecessor,
    prune,
    quote,
    reg_exp_js_match,
    reg_exp_js_replace,
    reg_exp_replace,
    repeat,
    replace,
    replace_end,
    replace_start,
    separator_case,
    series_phrase,
    series_phrase_serial,
    slugify,
    snake_case,
    split,
    start_case,
    starts_with,
    strip_tags,
    substr_left,
    substr_left_end,
    substr_right,
    substr_right_end,
    successor,
    surround,
    swap_case,
    title_case,
    to_lower,
    to_upper,
    trim,
    trim_end,
    trim_start,
    truncate,
    unescape,
    unquote,
    upper_case,
    upper_first,
    url,
    words,
)
from .utilities import (
    attempt,
    cond,
    conforms,
    conforms_to,
    constant,
    default_to,
    default_to_any,
    identity,
    iteratee,
    matches,
    matches_property,
    memoize,
    method,
    method_of,
    noop,
    now,
    nth_arg,
    over,
    over_every,
    over_some,
    properties,
    property_,
    property_of,
    random,
    range_,
    range_right,
    result,
    retry,
    stub_dict,
    stub_false,
    stub_list,
    stub_string,
    stub_true,
    times,
    to_path,
    unique_id,
)


py_ = _Dash()
_ = py_


__all__ = (
    "chunk",
    "compact",
    "concat",
    "difference",
    "difference_by",
    "difference_with",
    "drop",
    "drop_right",
    "drop_right_while",
    "drop_while",
    "duplicates",
    "fill",
    "find_index",
    "find_last_index",
    "flatten",
    "flatten_deep",
    "flatten_depth",
    "from_pairs",
    "head",
    "index_of",
    "initial",
    "intercalate",
    "interleave",
    "intersection",
    "intersection_by",
    "intersection_with",
    "intersperse",
    "last",
    "last_index_of",
    "mapcat",
    "nth",
    "pop",
    "pull",
    "pull_all",
    "pull_all_by",
    "pull_all_with",
    "pull_at",
    "push",
    "remove",
    "reverse",
    "shift",
    "slice_",
    "sort",
    "sorted_index",
    "sorted_index_by",
    "sorted_index_of",
    "sorted_last_index",
    "sorted_last_index_by",
    "sorted_last_index_of",
    "sorted_uniq",
    "sorted_uniq_by",
    "splice",
    "split_at",
    "tail",
    "take",
    "take_right",
    "take_right_while",
    "take_while",
    "union",
    "union_by",
    "union_with",
    "uniq",
    "uniq_by",
    "uniq_with",
    "unshift",
    "unzip",
    "unzip_with",
    "without",
    "xor",
    "xor_by",
    "xor_with",
    "zip_",
    "zip_object",
    "zip_object_deep",
    "zip_with",
    "_Dash",
    "chain",
    "tap",
    "thru",
    "at",
    "count_by",
    "every",
    "filter_",
    "find",
    "find_last",
    "flat_map",
    "flat_map_deep",
    "flat_map_depth",
    "for_each",
    "for_each_right",
    "group_by",
    "includes",
    "invoke_map",
    "key_by",
    "map_",
    "nest",
    "order_by",
    "partition",
    "pluck",
    "reduce_",
    "reduce_right",
    "reductions",
    "reductions_right",
    "reject",
    "sample",
    "sample_size",
    "shuffle",
    "size",
    "some",
    "sort_by",
    "InvalidMethod",
    "after",
    "ary",
    "before",
    "conjoin",
    "curry",
    "curry_right",
    "debounce",
    "delay",
    "disjoin",
    "flip",
    "flow",
    "flow_right",
    "iterated",
    "juxtapose",
    "negate",
    "once",
    "over_args",
    "partial",
    "partial_right",
    "rearg",
    "spread",
    "throttle",
    "unary",
    "wrap",
    "add",
    "ceil",
    "clamp",
    "divide",
    "floor",
    "max_",
    "max_by",
    "mean",
    "mean_by",
    "median",
    "min_",
    "min_by",
    "moving_mean",
    "multiply",
    "power",
    "round_",
    "scale",
    "slope",
    "std_deviation",
    "subtract",
    "sum_",
    "sum_by",
    "transpose",
    "variance",
    "zscore",
    "assign",
    "assign_with",
    "callables",
    "clone",
    "clone_deep",
    "clone_deep_with",
    "clone_with",
    "defaults",
    "defaults_deep",
    "find_key",
    "find_last_key",
    "for_in",
    "for_in_right",
    "get",
    "has",
    "invert",
    "invert_by",
    "invoke",
    "keys",
    "map_keys",
    "map_values",
    "map_values_deep",
    "merge",
    "merge_with",
    "omit",
    "omit_by",
    "parse_int",
    "pick",
    "pick_by",
    "rename_keys",
    "set_",
    "set_with",
    "to_boolean",
    "to_dict",
    "to_integer",
    "to_list",
    "to_number",
    "to_pairs",
    "to_string",
    "transform",
    "unset",
    "update",
    "update_with",
    "values",
    "eq",
    "gt",
    "gte",
    "in_range",
    "is_associative",
    "is_blank",
    "is_boolean",
    "is_builtin",
    "is_date",
    "is_decreasing",
    "is_dict",
    "is_empty",
    "is_equal",
    "is_equal_with",
    "is_error",
    "is_even",
    "is_float",
    "is_function",
    "is_increasing",
    "is_indexed",
    "is_instance_of",
    "is_integer",
    "is_iterable",
    "is_json",
    "is_list",
    "is_match",
    "is_match_with",
    "is_monotone",
    "is_nan",
    "is_negative",
    "is_none",
    "is_number",
    "is_object",
    "is_odd",
    "is_positive",
    "is_reg_exp",
    "is_set",
    "is_strictly_decreasing",
    "is_strictly_increasing",
    "is_string",
    "is_tuple",
    "is_zero",
    "lt",
    "lte",
    "camel_case",
    "capitalize",
    "chars",
    "chop",
    "chop_right",
    "clean",
    "count_substr",
    "deburr",
    "decapitalize",
    "ends_with",
    "ensure_ends_with",
    "ensure_starts_with",
    "escape",
    "escape_reg_exp",
    "has_substr",
    "human_case",
    "insert_substr",
    "join",
    "kebab_case",
    "lines",
    "lower_case",
    "lower_first",
    "number_format",
    "pad",
    "pad_end",
    "pad_start",
    "pascal_case",
    "predecessor",
    "prune",
    "quote",
    "reg_exp_js_match",
    "reg_exp_js_replace",
    "reg_exp_replace",
    "repeat",
    "replace",
    "replace_end",
    "replace_start",
    "separator_case",
    "series_phrase",
    "series_phrase_serial",
    "slugify",
    "snake_case",
    "split",
    "start_case",
    "starts_with",
    "strip_tags",
    "substr_left",
    "substr_left_end",
    "substr_right",
    "substr_right_end",
    "successor",
    "surround",
    "swap_case",
    "title_case",
    "to_lower",
    "to_upper",
    "trim",
    "trim_end",
    "trim_start",
    "truncate",
    "unescape",
    "unquote",
    "upper_case",
    "upper_first",
    "url",
    "words",
    "attempt",
    "cond",
    "conforms",
    "conforms_to",
    "constant",
    "default_to",
    "default_to_any",
    "identity",
    "iteratee",
    "matches",
    "matches_property",
    "memoize",
    "method",
    "method_of",
    "noop",
    "now",
    "nth_arg",
    "over",
    "over_every",
    "over_some",
    "properties",
    "property_",
    "property_of",
    "random",
    "range_",
    "range_right",
    "result",
    "retry",
    "stub_dict",
    "stub_false",
    "stub_list",
    "stub_string",
    "stub_true",
    "times",
    "to_path",
    "unique_id",
)



import typing as t

import pytest

import pydash as _


@pytest.mark.mypy_testing
def test_mypy_chunk() -> None:
    reveal_type(_.chunk([1, 2, 3, 4, 5], 2))  # R: builtins.list[typing.Sequence[builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_compact() -> None:
    reveal_type(_.compact([True, False, None]))  # R: builtins.list[builtins.bool]
    reveal_type(_.compact(['', 'hello', None]))  # R: builtins.list[builtins.str]
    reveal_type(_.compact([0, 1, None]))  # R: builtins.list[builtins.int]
    reveal_type(_.compact([0, 1]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_concat() -> None:
    reveal_type(_.concat([1, 2], [3, 4]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_difference() -> None:
    reveal_type(_.difference([1, 2, 3], [1], [2]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_difference_by() -> None:
    reveal_type(_.difference_by([1.2, 1.5, 1.7, 2.8], [0.9, 3.2], round))  # R: builtins.list[builtins.float]

    reveal_type(_.difference_by([{"hello": 1}], [{"hello": 2}], lambda d: d["hello"]))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_difference_with() -> None:
    array = ['apple', 'banana', 'pear']
    others = (['avocado', 'pumpkin'], ['peach'])
    def comparator(a: str, b: str) -> bool:
        return a[0] == b[0]

    reveal_type(_.difference_with(array, *others, comparator=comparator))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_drop() -> None:
    reveal_type(_.drop([1, 2, 3, 4], 2))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_drop_right() -> None:
    reveal_type(_.drop_right([1, 2, 3, 4], 2))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_drop_right_while() -> None:
    reveal_type(_.drop_right_while([1, 2, 3, 4], lambda x: x >= 3))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_drop_while() -> None:
    reveal_type(_.drop_while([1, 2, 3, 4], lambda x: x < 3))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_duplicates() -> None:
    reveal_type(_.duplicates([0, 1, 3, 2, 3, 1]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_fill() -> None:
    reveal_type(_.fill([1, 2, 3, 4, 5], 0))  # R: builtins.list[builtins.int]
    reveal_type(_.fill([1, 2, 3, 4, 5], 0, 1, 3))  # R: builtins.list[builtins.int]
    reveal_type(_.fill([1, 2, 3, 4, 5], 0, 0, 100))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_find_index() -> None:
    reveal_type(_.find_index([1, 2, 3, 4], lambda x: x >= 3))  # R: builtins.int
    reveal_type(_.find_index([1, 2, 3, 4], lambda x: x > 4))  # R: builtins.int
    reveal_type(_.find_index([{"a": 0, "b": 3}, {"a": "1", "c": 5}], {"a": 0}))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_find_last_index() -> None:
    reveal_type(_.find_last_index([1, 2, 3, 4], lambda x: x >= 3))  # R: builtins.int
    reveal_type(_.find_last_index([1, 2, 3, 4], lambda x: x > 4))  # R: builtins.int
    reveal_type(_.find_last_index([{"a": 0, "b": 3}, {"a": "1", "c": 5}], {"a": 0}))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_flatten() -> None:
    my_list: t.List[t.List[t.Union[int, t.List[int]]]] = [[1], [2, [3]], [[4]]]
    reveal_type(_.flatten(my_list))  # R: builtins.list[Union[builtins.int, builtins.list[builtins.int]]]


@pytest.mark.mypy_testing
def test_mypy_flatten_deep() -> None:
    reveal_type(_.flatten_deep([[1], [2, [3]], [[4]]]))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_flatten_depth() -> None:
    reveal_type(_.flatten_depth([[[1], [2, [3]], [[4]]]], 1))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_from_pairs() -> None:
    my_list: t.List[t.List[t.Union[str, int]]] = [['a', 1], ['b', 2]]
    reveal_type(_.from_pairs(my_list))  # R: builtins.dict[Union[builtins.str, builtins.int], Union[builtins.str, builtins.int]]

    my_list2: t.List[t.Tuple[str, int]] = [('a', 1), ('b', 2)]
    reveal_type(_.from_pairs(my_list2))  # R: builtins.dict[builtins.str, builtins.int]
#
#
@pytest.mark.mypy_testing
def test_mypy_head() -> None:
    reveal_type(_.head([1, 2, 3, 4]))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_index_of() -> None:
    reveal_type(_.index_of([1, 2, 3, 4], 2))  # R: builtins.int
    reveal_type(_.index_of([2, 1, 2, 3], 2, from_index=1))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_initial() -> None:
    reveal_type(_.initial([1, 2, 3, 4]))  # R: typing.Sequence[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_intercalate() -> None:
    my_list: t.List[t.List[int]] = [[2], [3]]
    reveal_type(_.intercalate(my_list, 'x'))  # R: builtins.list[Union[builtins.int, builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_interleave() -> None:
    reveal_type(_.interleave([1, 2, 3], [4, 5, 6], [7, 8, 9]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_intersection() -> None:
    reveal_type(_.intersection([1, 2, 3], [1, 2, 3, 4, 5], [2, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.intersection([1, 2, 3]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_intersection_by() -> None:
    reveal_type(_.intersection_by([1.2, 1.5, 1.7, 2.8], [0.9, 3.2], round))  # R: builtins.list[builtins.float]

    reveal_type(_.intersection_by([{"hello": 1}], [{"hello": 2}], lambda d: d["hello"]))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_intersection_with() -> None:
    array = ['apple', 'banana', 'pear']
    others = (['avocado', 'pumpkin'], ['peach'])
    def comparator(a: str, b:str) -> bool:
        return a[0] == b[0]

    reveal_type(_.intersection_with(array, *others, comparator=comparator))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_intersperse() -> None:
    my_list: t.List[t.Union[int, t.List[int]]] = [1, [2], [3], 4]
    reveal_type(_.intersperse(my_list, 'x'))  # R: builtins.list[Union[builtins.int, builtins.list[builtins.int], builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_last() -> None:
    reveal_type(_.last([1, 2, 3, 4]))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_last_index_of() -> None:
    reveal_type(_.last_index_of([1, 2, 2, 4], 2))  # R: builtins.int
    reveal_type(_.last_index_of([1, 2, 2, 4], 2, from_index=1))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_mapcat() -> None:
    def to_list(x: int) -> t.List[int]:
        return list(range(x))

    reveal_type(_.mapcat(list(range(4)), to_list))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_nth() -> None:
    reveal_type(_.nth([1, 2, 3], 0))  # R: Union[builtins.int, None]
    reveal_type(_.nth([11, 22, 33]))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_pop() -> None:
    array = [1, 2, 3, 4]
    reveal_type(_.pop(array))  # R: builtins.int
    reveal_type(_.pop(array, index=0))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_pull() -> None:
    reveal_type(_.pull([1, 2, 2, 3, 3, 4], 2, 3))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_pull_all() -> None:
    reveal_type(_.pull_all([1, 2, 2, 3, 3, 4], [2, 3]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_pull_all_by() -> None:
    array = [{'x': 1}, {'x': 2}, {'x': 3}, {'x': 1}]
    reveal_type(_.pull_all_by(array, [{'x': 1}, {'x': 3}], 'x'))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_pull_all_with() -> None:
    array = [{'x': 1, 'y': 2}, {'x': 3, 'y': 4}, {'x': 5, 'y': 6}]
    reveal_type(_.pull_all_with(array, [{'x': 3, 'y': 4}], lambda a, b: a == b))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]
    reveal_type(_.pull_all_with(array, [{'x': 3, 'y': 4}], lambda a, b: a != b))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_pull_at() -> None:
    reveal_type(_.pull_at([1, 2, 3, 4], 0, 2))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_push() -> None:
    array = [1, 2, 3]
    reveal_type(_.push(array, [4], [6]))  # R: builtins.list[Union[builtins.int, builtins.list[builtins.int]]]


@pytest.mark.mypy_testing
def test_mypy_remove() -> None:
    reveal_type(_.remove([1, 2, 3, 4], lambda x: x >= 3))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_reverse() -> None:
    reveal_type(_.reverse([1, 2, 3, 4]))  # R: builtins.list[builtins.int]
    reveal_type(_.reverse("hello"))  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_shift() -> None:
    array = [1, 2, 3, 4]
    reveal_type(_.shift(array))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_slice_() -> None:
    reveal_type(_.slice_([1, 2, 3, 4]))  # R: builtins.list[builtins.int]
    reveal_type(_.slice_([1, 2, 3, 4], 1))  # R: builtins.list[builtins.int]
    reveal_type(_.slice_([1, 2, 3, 4], 1, 3))  # R: builtins.list[builtins.int]
    reveal_type(_.slice_("hello", 1, 3))  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_sort() -> None:
    reveal_type(_.sort([2, 1, 4, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.sort([2, 1, 4, 3], reverse=True))  # R: builtins.list[builtins.int]

    value = _.sort(
        [{'a': 2, 'b': 1}, {'a': 3, 'b': 2}, {'a': 0, 'b': 3}],
        key=lambda item: item['a']
    )
    reveal_type(value)  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_sorted_index() -> None:
    reveal_type(_.sorted_index([1, 2, 2, 3, 4], 2))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_index_by() -> None:
    array = [{'x': 4}, {'x': 5}]
    reveal_type(_.sorted_index_by(array, {'x': 4}, lambda o: o['x']))  # R: builtins.int
    reveal_type(_.sorted_index_by(array, {'x': 4}, 'x'))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_index_of() -> None:
    reveal_type(_.sorted_index_of([3, 5, 7, 10], 3))  # R: builtins.int
    reveal_type(_.sorted_index_of([10, 10, 5, 7, 3], 10))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_last_index() -> None:
    reveal_type(_.sorted_last_index([1, 2, 2, 3, 4], 2))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_last_index_by() -> None:
    array = [{'x': 4}, {'x': 5}]
    reveal_type(_.sorted_last_index_by(array, {'x': 4}, lambda o: o['x']))  # R: builtins.int
    reveal_type(_.sorted_last_index_by(array, {'x': 4}, 'x'))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_last_index_of() -> None:
    reveal_type(_.sorted_last_index_of([4, 5, 5, 5, 6], 5))  # R: builtins.int
    reveal_type(_.sorted_last_index_of([6, 5, 5, 5, 4], 6))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sorted_uniq() -> None:
    reveal_type(_.sorted_uniq([4, 2, 2, 5]))  # R: builtins.list[builtins.int]
    reveal_type(_.sorted_uniq([-2, -2, 4, 1]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_sorted_uniq_by() -> None:
    reveal_type(_.sorted_uniq_by([3, 2, 1, 3, 2, 1], lambda val: val % 2))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_splice() -> None:
    array = [1, 2, 3, 4]
    reveal_type(_.splice(array, 1))  # R: builtins.list[builtins.int]

    array = [1, 2, 3, 4]
    reveal_type(_.splice(array, 1, 2))  # R: builtins.list[builtins.int]

    reveal_type(_.splice(array, 1, 2, 0, 0))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_split_at() -> None:
    reveal_type(_.split_at([1, 2, 3, 4], 2))  # R: builtins.list[typing.Sequence[builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_tail() -> None:
    reveal_type(_.tail([1, 2, 3, 4]))  # R: typing.Sequence[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_take() -> None:
    reveal_type(_.take([1, 2, 3, 4], 2))  # R: typing.Sequence[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_take_right() -> None:
    reveal_type(_.take_right([1, 2, 3, 4], 2))  # R: typing.Sequence[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_take_right_while() -> None:
    reveal_type(_.take_right_while([1, 2, 3, 4], lambda x: x >= 3))  # R: typing.Sequence[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_take_while() -> None:
    reveal_type(_.take_while([1, 2, 3, 4], lambda x: x < 3))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_union() -> None:
    reveal_type(_.union([1, 2, 3], [2, 3, 4], [3, 4, 5]))  # R: builtins.list[builtins.int]
    reveal_type(_.union([1, 2, 3], ["hello"]))  # R: builtins.list[Union[builtins.int, builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_union_by() -> None:
    reveal_type(_.union_by([1, 2, 3], [2, 3, 4], iteratee=lambda x: x % 2))  # R: builtins.list[builtins.int]

    reveal_type(_.union_by([{"hello": 1}], [{"hello": 2}], lambda d: d["hello"]))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_union_with() -> None:
    comparator = lambda a, b: (a % 2) == (b % 2)
    reveal_type(_.union_with([1, 2, 3], [2, 3, 4], comparator=comparator))  # R: builtins.list[builtins.int]
    reveal_type(_.union_with([1, 2, 3], [2, 3, 4]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_uniq() -> None:
    reveal_type(_.uniq([1, 2, 3, 1, 2, 3]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_uniq_by() -> None:
    reveal_type(_.uniq_by([1, 2, 3, 1, 2, 3], lambda val: val % 2))  # R: builtins.list[builtins.int]
    reveal_type(_.uniq_by([{"hello": 1}, {"hello": 1}], lambda val: val["hello"]))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_uniq_with() -> None:
    reveal_type(_.uniq_with([1, 2, 3, 4, 5], lambda a, b: (a % 2) == (b % 2)))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_unshift() -> None:
    array = [1, 2, 3, 4]
    reveal_type(_.unshift(array, -1, -2))  # R: builtins.list[builtins.int]
    reveal_type(_.unshift(array, "hello"))  # R: builtins.list[Union[builtins.int, builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_unzip() -> None:
    reveal_type(_.unzip([[1, 4, 7], [2, 5, 8], [3, 6, 9]]))  # R: builtins.list[builtins.list[builtins.int]]
    reveal_type(_.unzip([(1, 4, 7), (2, 5, 8), (3, 6, 9)]))  # R: builtins.list[builtins.list[builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_unzip_with() -> None:
    def add(x: int, y: int) -> int:
        return x + y

    reveal_type(_.unzip_with([[1, 10, 100], [2, 20, 200]], add))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_without() -> None:
    reveal_type(_.without([1, 2, 3, 2, 4, 4], 2, 4))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_xor() -> None:
    reveal_type(_.xor([1, 3, 4], [1, 2, 4], [2]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_xor_by() -> None:
    reveal_type(_.xor_by([2.1, 1.2], [2.3, 3.4], round))  # R: builtins.list[builtins.float]
    reveal_type(_.xor_by([{'x': 1}], [{'x': 2}, {'x': 1}], iteratee='x'))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]

    reveal_type(_.xor_by([{"hello": 1}], [{"hello": 2}], lambda d: d["hello"]))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_xor_with() -> None:
    objects = [{'x': 1, 'y': 2}, {'x': 2, 'y': 1}]
    others = [{'x': 1, 'y': 1}, {'x': 1, 'y': 2}]
    reveal_type(_.xor_with(objects, others, lambda a, b: a == b))  # R: builtins.list[builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_zip_() -> None:
    reveal_type(_.zip_([1, 2, 3], [4, 5, 6], [7, 8, 9]))  # R: builtins.list[builtins.list[builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_zip_object() -> None:
    reveal_type(_.zip_object([1, 2, 3], [4, 5, 6]))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.zip_object([1, 2, 3], ["hello", "good", "friend"]))  # R: builtins.dict[builtins.int, builtins.str]

    my_list: t.List[t.Tuple[int, str]] = [(1, "hello"), (2, "good"), (3, "friend")]
    reveal_type(_.zip_object(my_list))  # R: builtins.dict[builtins.int, builtins.str]


@pytest.mark.mypy_testing
def test_mypy_zip_object_deep() -> None:
    reveal_type(_.zip_object_deep(['a.b.c', 'a.b.d'], [1, 2]))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_zip_with() -> None:
    def add(x: int, y: int) -> int:
        return x + y

    reveal_type(_.zip_with([1, 2], [10, 20], [100, 200], add))  # R: builtins.list[Union[builtins.list[builtins.int], builtins.int]]
    reveal_type(_.zip_with([1, 2], [10, 20], [100, 200], iteratee=add))  # R: builtins.list[builtins.int]
